@@ -46,6 +46,7 @@ struct connman_device {
 	char *node;
 	char *address;
 	char *interface;
+	char *control;
 	char *ident;
 	unsigned int connections;
 	guint scan_timeout;
@@ -192,12 +193,43 @@ static int set_carrier(struct connman_device *device, connman_bool_t carrier)
 {
 	struct connman_service *service;
 
+	if (carrier == TRUE)
+		__connman_profile_add_device(device);
+	else
+		__connman_profile_remove_device(device);
+
 	service = __connman_service_lookup_from_device(device);
 	__connman_service_set_carrier(service, carrier);
 
 	return set_connected(device, carrier);
 }
 
+static int powered_changed(struct connman_device *device)
+{
+	DBusMessage *signal;
+	DBusMessageIter entry, value;
+	const char *key = "Powered";
+
+	signal = dbus_message_new_signal(device->element.path,
+				CONNMAN_DEVICE_INTERFACE, "PropertyChanged");
+	if (signal == NULL)
+		return -ENOMEM;
+
+	dbus_message_iter_init_append(signal, &entry);
+
+	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
+
+	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
+					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
+	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN,
+							&device->powered);
+	dbus_message_iter_close_container(&entry, &value);
+
+	g_dbus_send_message(connection, signal);
+
+	return 0;
+}
+
 static int set_powered(struct connman_device *device, connman_bool_t powered)
 {
 	struct connman_device_driver *driver = device->driver;
@@ -205,6 +237,9 @@ static int set_powered(struct connman_device *device, connman_bool_t powered)
 
 	DBG("device %p powered %d", device, powered);
 
+	if (device->powered == powered)
+		return -EALREADY;
+
 	if (!driver)
 		return -EINVAL;
 
@@ -226,6 +261,11 @@ static int set_powered(struct connman_device *device, connman_bool_t powered)
 			err = -EINVAL;
 	}
 
+	if (err == 0) {
+		device->powered = powered;
+		powered_changed(device);
+	}
+
 	return err;
 }
 
@@ -456,26 +496,6 @@ static DBusMessage *set_property(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static char *build_group(const unsigned char *ssid, unsigned int ssid_len,
-					const char *mode, const char *security)
-{
-	GString *str;
-	unsigned int i;
-
-	str = g_string_sized_new((ssid_len * 2) + 24);
-	if (str == NULL)
-		return NULL;
-
-	if (ssid_len > 0 && ssid[0] != '\0') {
-		for (i = 0; i < ssid_len; i++)
-			g_string_append_printf(str, "%02x", ssid[i]);
-	}
-
-	g_string_append_printf(str, "_%s_%s", mode, security);
-
-	return g_string_free(str, FALSE);
-}
-
 static void convert_name(const char *ssid, char *name,
 						unsigned int ssid_len)
 {
@@ -634,7 +654,7 @@ static DBusMessage *join_network(DBusConnection *conn,
 		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 	}
 
-	group = build_group(ssid, ssid_size, mode, security);
+	group = connman_wifi_build_group_name(ssid, ssid_size, mode, security);
 
 	index = connman_device_get_index(device);
 	connman_network_set_index(network, index);
@@ -695,13 +715,7 @@ static DBusMessage *propose_scan(DBusConnection *conn,
 		break;
 	}
 
-	if (!device->driver || !device->driver->scan)
-		return __connman_error_not_supported(msg);
-
-	if (device->powered == FALSE)
-		return __connman_error_failed(msg, EINVAL);
-
-	err = device->driver->scan(device);
+	err = __connman_device_scan(device);
 	if (err < 0)
 		return __connman_error_failed(msg, -err);
 
@@ -803,7 +817,8 @@ static void device_enable(struct connman_device *device)
 		return;
 
 	if (device->driver->enable) {
-		device->driver->enable(device);
+		if (device->driver->enable(device) == 0)
+			device->powered = TRUE;
 		__connman_notifier_device_type_increase(device->type);
 	}
 }
@@ -818,7 +833,8 @@ static void device_disable(struct connman_device *device)
 	g_hash_table_remove_all(device->networks);
 
 	if (device->driver->disable) {
-		device->driver->disable(device);
+		if (device->driver->disable(device) == 0)
+			device->powered = FALSE;
 		__connman_notifier_device_type_decrease(device->type);
 	}
 }
@@ -843,7 +859,7 @@ static int setup_device(struct connman_device *device)
 	case CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE:
 		break;
 	case CONNMAN_DEVICE_MODE_TRANSPORT_IP:
-		if (device->secondary == FALSE)
+		if (device->carrier == TRUE && device->secondary == FALSE)
 			__connman_profile_add_device(device);
 		break;
 	}
@@ -1000,6 +1016,7 @@ static void device_destruct(struct connman_element *element)
 	g_free(device->node);
 	g_free(device->name);
 	g_free(device->address);
+	g_free(device->control);
 	g_free(device->interface);
 
 	connman_ipconfig_unref(device->ipconfig);
@@ -1182,11 +1199,12 @@ int connman_device_get_index(struct connman_device *device)
  * connman_device_set_interface:
  * @device: device structure
  * @interface: interface name
+ * @control: control interface
  *
  * Set interface name of device
  */
 void connman_device_set_interface(struct connman_device *device,
-							const char *interface)
+				const char *interface, const char *control)
 {
 	g_free(device->element.devname);
 	device->element.devname = g_strdup(interface);
@@ -1194,6 +1212,9 @@ void connman_device_set_interface(struct connman_device *device,
 	g_free(device->interface);
 	device->interface = g_strdup(interface);
 
+	g_free(device->control);
+	device->control = g_strdup(control);
+
 	if (device->name == NULL) {
 		const char *str = type2description(device->type);
 		if (str != NULL && device->interface != NULL)
@@ -1202,15 +1223,9 @@ void connman_device_set_interface(struct connman_device *device,
 	}
 }
 
-/**
- * connman_device_get_interface:
- * @device: device structure
- *
- * Get interface name of device
- */
-const char *connman_device_get_interface(struct connman_device *device)
+const char *connman_device_get_control(struct connman_device *device)
 {
-	return device->interface;
+	return device->control;
 }
 
 /**
@@ -1264,7 +1279,7 @@ enum connman_device_mode connman_device_get_mode(struct connman_device *device)
  * Change secondary value of device
  */
 void connman_device_set_secondary(struct connman_device *device,
-                                                connman_bool_t secondary)
+						connman_bool_t secondary)
 {
 	device->secondary = secondary;
 }
@@ -1290,10 +1305,6 @@ connman_bool_t connman_device_get_secondary(struct connman_device *device)
 int connman_device_set_powered(struct connman_device *device,
 						connman_bool_t powered)
 {
-	DBusMessage *signal;
-	DBusMessageIter entry, value;
-	const char *key = "Powered";
-
 	DBG("driver %p powered %d", device, powered);
 
 	if (device->timeout > 0) {
@@ -1317,21 +1328,7 @@ int connman_device_set_powered(struct connman_device *device,
 	if (device->registered == FALSE)
 		return 0;
 
-	signal = dbus_message_new_signal(device->element.path,
-				CONNMAN_DEVICE_INTERFACE, "PropertyChanged");
-	if (signal == NULL)
-		return 0;
-
-	dbus_message_iter_init_append(signal, &entry);
-
-	dbus_message_iter_append_basic(&entry, DBUS_TYPE_STRING, &key);
-
-	dbus_message_iter_open_container(&entry, DBUS_TYPE_VARIANT,
-					DBUS_TYPE_BOOLEAN_AS_STRING, &value);
-	dbus_message_iter_append_basic(&value, DBUS_TYPE_BOOLEAN, &powered);
-	dbus_message_iter_close_container(&entry, &value);
-
-	g_dbus_send_message(connection, signal);
+	powered_changed(device);
 
 	if (powered == FALSE)
 		return 0;
@@ -1382,6 +1379,17 @@ int connman_device_set_carrier(struct connman_device *device,
 	return set_carrier(device, device->carrier);
 }
 
+int __connman_device_scan(struct connman_device *device)
+{
+	if (!device->driver || !device->driver->scan)
+		return -EOPNOTSUPP;
+
+	if (device->powered == FALSE)
+		return -ENOLINK;
+
+	return device->driver->scan(device);
+}
+
 int __connman_device_connect(struct connman_device *device)
 {
 	DBG("device %p", device);
@@ -1409,7 +1417,8 @@ int __connman_device_disconnect(struct connman_device *device)
 	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
 		struct connman_network *network = value;
 
-		if (connman_network_get_connected(network) == TRUE)
+		if (connman_network_get_connected(network) == TRUE ||
+				connman_network_get_associating(network) == TRUE)
 			__connman_network_disconnect(network);
 	}
 
@@ -1789,6 +1798,8 @@ int connman_device_register(struct connman_device *device)
 		break;
 	}
 
+	__connman_notifier_device_type_register(device->type);
+
 	return connman_element_register(&device->element, NULL);
 }
 
@@ -1802,6 +1813,8 @@ void connman_device_unregister(struct connman_device *device)
 {
 	__connman_storage_save_device(device);
 
+	__connman_notifier_device_type_unregister(device->type);
+
 	connman_element_unregister(&device->element);
 }
 
