@@ -34,7 +34,6 @@ struct connman_network {
 	struct connman_element element;
 	enum connman_network_type type;
 	enum connman_network_protocol protocol;
-	connman_bool_t associating;
 	connman_bool_t secondary;
 	connman_bool_t available;
 	connman_bool_t connected;
@@ -51,6 +50,8 @@ struct connman_network {
 	void *driver_data;
 
 	connman_bool_t registered;
+	connman_bool_t connecting;
+	connman_bool_t associating;
 
 	struct connman_device *device;
 
@@ -77,6 +78,7 @@ static const char *type2string(enum connman_network_type type)
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
 		return "bluetooth";
+	case CONNMAN_NETWORK_TYPE_MBM:
 	case CONNMAN_NETWORK_TYPE_HSO:
 		return "cellular";
 	}
@@ -610,8 +612,9 @@ void connman_network_set_group(struct connman_network *network,
 	case CONNMAN_NETWORK_TYPE_VENDOR:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
-	case CONNMAN_NETWORK_TYPE_HSO:
 		return;
+	case CONNMAN_NETWORK_TYPE_MBM:
+	case CONNMAN_NETWORK_TYPE_HSO:
 	case CONNMAN_NETWORK_TYPE_WIFI:
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 		break;
@@ -648,6 +651,29 @@ const char *__connman_network_get_ident(struct connman_network *network)
 	return __connman_device_get_ident(network->device);
 }
 
+connman_bool_t __connman_network_get_weakness(struct connman_network *network)
+{
+	if (network->secondary == TRUE)
+		return FALSE;
+
+	switch (network->type) {
+	case CONNMAN_NETWORK_TYPE_UNKNOWN:
+	case CONNMAN_NETWORK_TYPE_VENDOR:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_MBM:
+	case CONNMAN_NETWORK_TYPE_HSO:
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+		break;
+	case CONNMAN_NETWORK_TYPE_WIFI:
+		if (network->strength < 20)
+			return TRUE;
+		break;
+	}
+
+	return FALSE;
+}
+
 /**
  * connman_network_set_available:
  * @network: network structure
@@ -758,6 +784,9 @@ static gboolean set_connected(gpointer user_data)
 						CONNMAN_SERVICE_STATE_IDLE);
 	}
 
+	network->connecting = FALSE;
+	network->associating = FALSE;
+
 	return FALSE;
 }
 
@@ -777,14 +806,23 @@ int connman_network_set_connected(struct connman_network *network,
 
 	DBG("network %p connected %d", network, connected);
 
+	if ((network->connecting == TRUE || network->associating == TRUE) &&
+							connected == FALSE) {
+		connman_element_set_error(&network->element,
+					CONNMAN_ELEMENT_ERROR_CONNECT_FAILED);
+		__connman_network_disconnect(network);
+	}
+
+	if (connected == TRUE) {
+		network->connecting = FALSE;
+		network->associating = FALSE;
+	}
+
 	if (network->connected == connected)
 		return -EALREADY;
 
 	network->connected = connected;
 
-	if (connected == TRUE)
-		network->associating = FALSE;
-
 	if (network->registered == FALSE) {
 		g_idle_add(set_connected, network);
 		return 0;
@@ -822,6 +860,17 @@ connman_bool_t connman_network_get_connected(struct connman_network *network)
 	return network->connected;
 }
 
+/**
+ * connman_network_get_associating:
+ * @network: network structure
+ *
+ * Get network associating status
+ */
+connman_bool_t connman_network_get_associating(struct connman_network *network)
+{
+	return network->associating;
+}
+
 /**
  * __connman_network_connect:
  * @network: network structure
@@ -835,6 +884,9 @@ int __connman_network_connect(struct connman_network *network)
 	DBG("network %p", network);
 
 	if (network->connected == TRUE)
+		return -EISCONN;
+
+	if (network->connecting == TRUE || network->associating == TRUE)
 		return -EALREADY;
 
 	if (network->driver == NULL)
@@ -845,12 +897,19 @@ int __connman_network_connect(struct connman_network *network)
 
 	__connman_device_disconnect(network->device);
 
+	network->connecting = TRUE;
+
 	err = network->driver->connect(network);
-	if (err == 0) {
-		network->connected = TRUE;
-		set_connected(network);
+	if (err < 0) {
+		if (err == -EINPROGRESS)
+			connman_network_set_associating(network, TRUE);
+
+		return err;
 	}
 
+	network->connected = TRUE;
+	set_connected(network);
+
 	return err;
 }
 
@@ -866,7 +925,8 @@ int __connman_network_disconnect(struct connman_network *network)
 
 	DBG("network %p", network);
 
-	if (network->connected == FALSE)
+	if (network->connected == FALSE && network->connecting == FALSE &&
+						network->associating == FALSE)
 		return -ENOTCONN;
 
 	if (network->driver == NULL)
@@ -875,6 +935,8 @@ int __connman_network_disconnect(struct connman_network *network)
 	if (network->driver->disconnect == NULL)
 		return -ENOSYS;
 
+	network->connecting = FALSE;
+
 	err = network->driver->disconnect(network);
 	if (err == 0) {
 		network->connected = FALSE;
@@ -1238,8 +1300,9 @@ static int network_probe(struct connman_element *element)
 	case CONNMAN_NETWORK_TYPE_VENDOR:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
-	case CONNMAN_NETWORK_TYPE_HSO:
 		break;
+	case CONNMAN_NETWORK_TYPE_MBM:
+	case CONNMAN_NETWORK_TYPE_HSO:
 	case CONNMAN_NETWORK_TYPE_WIFI:
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 		if (network->group != NULL && network->secondary == FALSE)
@@ -1267,8 +1330,9 @@ static void network_remove(struct connman_element *element)
 	case CONNMAN_NETWORK_TYPE_VENDOR:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
-	case CONNMAN_NETWORK_TYPE_HSO:
 		break;
+	case CONNMAN_NETWORK_TYPE_MBM:
+	case CONNMAN_NETWORK_TYPE_HSO:
 	case CONNMAN_NETWORK_TYPE_WIFI:
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 		if (network->group != NULL && network->secondary == FALSE) {
